unit uCtDBD;

(*
  This unit contains the TCtTableEditor class.  This class is used to restructure existing tables.  The basic
  process is as follows :
    1. Register all the fields that you would like your table to have after the restructure
    2. Register the table that will be restructured
    3. Execute the restructure
  To create a new table the process is this :
    1. Register all the fields that you would like your table to have
    2. Create the table

  Necessary additions to TCtTableEditor :
  Register indexes.

  There is no way to implicitly rename a field.  If a rename is required is could be done
  after registering the table but before executing the restructure.  This is what would have to happen :
    1. Find the field number in the original table
    2. Find the field number in the new table
    3. pFieldDesc^.iFldNum := original field number
    4. pFieldOps^ := crMODIFY
*)

interface

uses
  SysUtils, DB, DBTables, BDE, Grids, Dialogs,
  Controls, StdCtrls;

type
  TCtTableEditor = class
    // table properties
    FTable: TTable;
    FFieldDefs: TFieldDefs;
    FIndexDefs: TIndexDefs;
    FIsPackTable: Boolean;
    FKeyFieldCount: Integer;
    //
    pFieldDesc: pFLDDesc;
    pFieldOps: pCROpType;
    //
    pIndexDesc: pIDXDesc;
    pIndexOps: pCROpType;
    //
    constructor Create;
    destructor Destroy; override;
    procedure Clear;
    procedure RegisterField(FieldName: String; FieldType: TFieldType; Size: Word);
    procedure RegisterIndex(const IndexName: string; const IndexFields: String; IndexOptions: TIndexOptions);
    procedure CreateTable(TblName: String);
    function RegisterTable(var Table: TTable): Boolean;
    function Execute: Boolean;
  end;

var
  BDEFieldTypeName: Array[ftUnknown..ftTypedBinary] of String;
  BDEFieldTypeSize: Array[ftUnknown..ftTypedBinary] of Boolean;
  PDX2BDEFieldType: Array[fldPDXCHAR..fldPDXUNICHAR] of TFieldType;

  // general library routines
  function ctGetIndex(Str: String; AR: Array of String): Integer;
  function ctStrToInt(Str: String): Integer;
  function ctIntToStr(I: Integer): String;
  procedure SetAutoSize(WinControl: TWinControl; Value: Boolean);

implementation

{ ---------------------------------------------------------
  Library Routines
  ---------------------------------------------------------}
function ctGetIndex(Str: String; AR: Array of String): Integer;
var
  i: Integer;
begin
  for i := low(AR) to high(AR) do if (0 = CompareText(Str, AR[i])) then break;
  if (i <= high(AR)) then result := i else result := -1;
end;

function ctStrToInt(Str:String):Integer;begin if(''=Str)then result:=0 else result:=StrToInt(Str);end;
function ctIntToStr(I: Integer):String;begin if(0=I)then result:=''else result:=IntToStr(I);end;

procedure SetAutoSize(WinControl: TWinControl; Value: Boolean);
var
  i: Integer;
  Control: TControl;
begin
  for i := 0 to WinControl.ControlCount-1 do begin
    Control := WinControl.Controls[i];
    if (Control is TLabel) then with TLabel(Control) do begin
      AutoSize := NOT(Value);
      AutoSize := Value;
    end;
    if (Control is TWinControl) then SetAutoSize(TWinControl(Control), Value);
  end;
end;

{ ---------------------------------------------------------
  TCtTableEditor
  Used to restructure a table.  The basic process is as follows :
  1. Register all the fields you want in the table
  2. Register the table that you want to restructure
  3. Perform the restructure
 ---------------------------------------------------------- }
constructor TCtTableEditor.Create;
begin
  FFieldDefs := TFieldDefs.Create(nil);
  FIndexDefs := TIndexDefs.Create(nil);
  pFieldDesc := nil;
  pFieldOps := nil;
  pIndexDesc := nil;
  pIndexOps := nil;
end;

(*
  Clean up.
*)
destructor TCtTableEditor.Destroy;
begin
  FFieldDefs.Free;
  FIndexDefs.Free;
  if (nil <> pFieldDesc) then freemem(pFieldDesc);
  if (nil <> pFieldOps) then freemem(pFieldOps);
  if (nil <> pIndexDesc) then freemem(pIndexDesc);
  if (nil <> pIndexOps) then freemem(pIndexOps);
end;

(*
  Register the fields that are wanted after the restructure.
*)
procedure TCtTableEditor.Clear;
begin
  FTable := nil;
  FIsPackTable := FALSE;
  FFieldDefs.Clear;
  FIndexDefs.Clear;
end;

(*
  Register the fields that are wanted after the restructure.
*)
procedure TCtTableEditor.RegisterField(FieldName: String; FieldType: TFieldType; Size: Word);
begin
  FFieldDefs.Add(FieldName, FieldType, Size, FALSE);
end;

(*
  Register the indexes that are wanted after the restructure.
*)
procedure TCtTableEditor.RegisterIndex(const IndexName: String; const IndexFields: String; IndexOptions: TIndexOptions);
begin
  ShowMessage('Indexes are not supported.');
end;

(*
  Save the table as a new table
*)
procedure TCtTableEditor.CreateTable(TblName: String);
var
  DB: TDatabase;
  Table: TTable;
begin
  DB := TDatabase.Create(nil);
  Table := TTable.Create(nil);
  try
    DB.Params.Add(format('PATH = "%s"', [ExtractFilePath(TblName)]));
    DB.Drivername := 'STANDARD';
    DB.DatabaseName := 'CTDBDTEMPDB'; // any unused name
    DB.Connected := TRUE;
    Table.TableType := ttParadox;
    Table.DatabaseName := DB.DatabaseName;
    Table.TableName := TblName;
    Table.FieldDefs.Assign(FFieldDefs);
    Table.IndexDefs.Assign(FIndexDefs);
    Table.CreateTable;
  finally
    DB.Free;
    Table.Free;
  end;
end;

(*
  Table must be opened exclusively.  Changes necessary to restructure table to
  reflect the registered structure are determined.

  Set the Field Ops to restructure a table.
  FldCount = Number of field in the new table
  OldFieldDesc = Original field description.
  NewFieldDesc = New field description.
  Ops = Field operation to perform.
*)
function TCtTableEditor.RegisterTable(var Table: TTable): Boolean;
var
  FField: TField;
  pField: pFLDDesc;
  pIndex: pIDXDesc;
  pOp: pCROpType;
  Props: CURProps;
  i, j: Integer;
begin
  FTable := Table;
  // Make sure the table is open exclusively so we can get the db handle...
  if NOT(Table.Active) then raise EDatabaseError.Create('Table must be opened to register');
  if NOT(Table.Exclusive) then raise EDatabaseError.Create('Table must be opened exclusively to register');
  // Make sure the table is either Paradox or dBASE...
  Check(DbiSetProp(hDBIObj(Table.Handle), curxltMODE, integer(xltNONE)));
  Check(DbiGetCursorProps(Table.Handle, Props));
  if NOT(Props.szTableType = szPARADOX) then raise EDatabaseError.Create('Only Paradox tables may be registered');
  // okay so far, lets see if any fields need to be added, modified or copied (moved).
  if (nil <> pFieldDesc) then freemem(pFieldDesc);
  if (nil <> pFieldOps) then freemem(pFieldOps);
  if (nil <> pIndexDesc) then freemem(pIndexDesc);
  if (nil <> pIndexOps) then freemem(pIndexOps);
  pFieldDesc := AllocMem(FFieldDefs.Count * sizeof(FLDDesc));
  pFieldOps := AllocMem(FFieldDefs.Count * sizeof(CROpType));
  pIndexDesc := AllocMem(FIndexDefs.Count * sizeof(IDXDesc));
  pIndexOps := AllocMem(FIndexDefs.Count * sizeof(CROpType));
  try
    FillChar(pFieldDesc^, FFieldDefs.Count * sizeof(FLDDesc), #0);
    FillChar(pFieldOps^, FFieldDefs.Count * sizeof(CROpType), #0);
    FillChar(pIndexDesc^, FIndexDefs.Count * sizeof(IDXDesc), #0);
    FillChar(pIndexOps^, FIndexDefs.Count * sizeof(CROpType), #0);
    // determine field changes
    pField := pFieldDesc;
    pOp := pFieldOps;
    for i := 0 to FFieldDefs.Count -1 do begin
      with pField^ do begin
        strcopy(szName, PChar(FFieldDefs[i].Name));
        for j := fldPDXCHAR to fldPDXUNICHAR do if (FFieldDefs[i].DataType = PDX2BDEFieldType[j]) then break;
        if (j > fldPDXUNICHAR) then raise EDatabaseError.Create(format('Invalid Datatype in field %s', [FFieldDefs[i].Name]));
        iFldType := j;
        if BDEFieldTypeSize[FFieldDefs[i].DataType] then iUnits1 := FFieldDefs[i].Size else iUnits1 := 1;
        FField := Table.FindField(FFieldDefs[i].Name);
        // if a new field then crADD else if type or size changed then crMODIFY else if moved then crCOPY.
        if (nil = FField) then begin
          iFldNum := 0;
          pOp^ := crADD;
        end else begin
          iFldNum := FField.FieldNo;
          if (FField.DataType <> FFieldDefs[i].DataType) OR (BDEFieldTypeSize[FFieldDefs[i].DataType]) AND (FField.Size <> FFieldDefs[i].Size) then pOp^ := crModify
          else if (iFldNum <> i+1) then pOp^ := crCOPY;
        end;
      end;
      inc(pField);
      inc(pOp);
    end;
    // currently works for the primary index only
    pIndex := pIndexDesc;
    pOp := pIndexOps;
    for i := 0 to FIndexDefs.Count -1 do begin
      FillChar(pIndex^, SizeOf(pIDXDesc), 0);
      with pIndex^ do begin
        strcopy(szName, 'Primary');
        bPrimary:= TRUE;
        bUnique:= TRUE;
        bMaintained:= TRUE;
        iFldsInKey:= FKeyFieldCount;
        for j := 1 to iFldsInKey do aiKeyFld[j-1] := j;
        bCaseInsensitive:= TRUE;
      end;
      if (0 < Table.IndexDefs.Count) then pOp^ := crMODIFY else pOp^ := crADD;
      inc(pIndex);
      inc(pOp);
    end;
    result := TRUE;
  except
    freemem(pFieldDesc); pFieldDesc := nil;
    freemem(pFieldOps); pFieldOps := nil;
    freemem(pIndexDesc); pIndexDesc := nil;
    freemem(pIndexOps); pIndexOps := nil;
    result := FALSE;
  end;
end;

(*
  Setup a table description and restructure the registered table.
*)
function TCtTableEditor.Execute: Boolean;
var
  TableDesc: CRTblDesc;
  hDb: hDbiDb;
begin
  try
    FillChar(TableDesc, sizeof(TableDesc), 0);
    Check(DbiGetObjFromObj(hDBIObj(FTable.Handle), objDATABASE, hDBIObj(hDb)));
    StrPCopy(TableDesc.szTblName, FTable.TableName);
    StrPCopy(TableDesc.szTblType, szPARADOX);
    //
    TableDesc.iFldCount := FFieldDefs.Count;
    TableDesc.pfldDesc := pFieldDesc;
    TableDesc.pecrFldOp := pFieldOps;
    //
    TableDesc.iIdxCount := FIndexDefs.Count;
    TableDesc.pIdxDesc := pIndexDesc;
    TableDesc.pecrIdxOp := pIndexOps;
    //
    TableDesc.bPack := FIsPackTable;
    FTable.Close;
    Check(DbiDoRestructure(hDb, 1, @TableDesc, nil, nil, nil, FALSE));
    result := TRUE;
  except
    result := FALSE;
  end;
end;

Initialization;
  BDEFieldTypeName[ftUnknown] := 'Unknown';
  BDEFieldTypeName[ftString] := 'String';
  BDEFieldTypeName[ftSmallint] := 'Smallint';
  BDEFieldTypeName[ftInteger] := 'Integer';
  BDEFieldTypeName[ftWord] := 'Word';
  BDEFieldTypeName[ftBoolean] := 'Boolean';
  BDEFieldTypeName[ftFloat] := 'Float';
  BDEFieldTypeName[ftCurrency] := 'Currency';
  BDEFieldTypeName[ftBCD] := 'BCD';
  BDEFieldTypeName[ftDate] := 'Date';
  BDEFieldTypeName[ftTime] := 'Time';
  BDEFieldTypeName[ftDateTime] := 'DateTime';
  BDEFieldTypeName[ftBytes] := 'Bytes';
  BDEFieldTypeName[ftVarBytes] := 'VarBytes';
  BDEFieldTypeName[ftAutoInc] := 'AutoInc';
  BDEFieldTypeName[ftBlob] := 'Blob';
  BDEFieldTypeName[ftMemo] := 'Memo';
  BDEFieldTypeName[ftGraphic] := 'Graphic';
  BDEFieldTypeName[ftFmtMemo] := 'FmtMemo';
  BDEFieldTypeName[ftParadoxOle] := 'ParadoxOle';
  BDEFieldTypeName[ftDBaseOle] := 'DBaseOle';
  BDEFieldTypeName[ftTypedBinary] := 'TypedBinary';
  //
  // note which data types need an explicit size
  BDEFieldTypeSize[ftUnknown] := TRUE;
  BDEFieldTypeSize[ftString] := TRUE;
  BDEFieldTypeSize[ftSmallint] := FALSE;
  BDEFieldTypeSize[ftInteger] := FALSE;
  BDEFieldTypeSize[ftWord] := FALSE;
  BDEFieldTypeSize[ftBoolean] := FALSE;
  BDEFieldTypeSize[ftFloat] := FALSE;
  BDEFieldTypeSize[ftCurrency] := FALSE;
  BDEFieldTypeSize[ftBCD] := TRUE;
  BDEFieldTypeSize[ftDate] := FALSE;
  BDEFieldTypeSize[ftTime] := FALSE;
  BDEFieldTypeSize[ftDateTime] := FALSE;
  BDEFieldTypeSize[ftBytes] := TRUE;
  BDEFieldTypeSize[ftVarBytes] := TRUE;
  BDEFieldTypeSize[ftAutoInc] := FALSE;
  BDEFieldTypeSize[ftBlob] := TRUE;
  BDEFieldTypeSize[ftMemo] := TRUE;
  BDEFieldTypeSize[ftGraphic] := TRUE;
  BDEFieldTypeSize[ftFmtMemo] := TRUE;
  BDEFieldTypeSize[ftParadoxOle] := TRUE;
  BDEFieldTypeSize[ftDBaseOle] := TRUE;
  BDEFieldTypeSize[ftTypedBinary] := TRUE;
  //
  PDX2BDEFieldType[fldPDXCHAR] := ftString;
  PDX2BDEFieldType[fldPDXNUM] := ftFloat;
  PDX2BDEFieldType[fldPDXMONEY] := ftCurrency;
  PDX2BDEFieldType[fldPDXDATE] := ftDate;
  PDX2BDEFieldType[fldPDXSHORT] := ftSmallint;
  PDX2BDEFieldType[fldPDXMEMO] := ftMemo;
  PDX2BDEFieldType[fldPDXBINARYBLOB] := ftBlob;
  PDX2BDEFieldType[fldPDXFMTMEMO] := ftFmtMemo;
  PDX2BDEFieldType[fldPDXOLEBLOB] := ftParadoxOle;
  PDX2BDEFieldType[fldPDXGRAPHIC] := ftGraphic;
  PDX2BDEFieldType[fldPDXBLOB] := ftBlob;
  PDX2BDEFieldType[fldPDXLONG] := ftInteger;
  PDX2BDEFieldType[fldPDXTIME] := ftTime;
  PDX2BDEFieldType[fldPDXDATETIME] := ftDateTime;
  PDX2BDEFieldType[fldPDXBOOL] := ftBoolean;
  PDX2BDEFieldType[fldPDXAUTOINC] := ftAutoInc;
  PDX2BDEFieldType[fldPDXBYTES] := ftBytes;
  PDX2BDEFieldType[fldPDXBCD] := ftBCD;
  //PDX2BDEFieldType[fldPDXUNICHAR] := 0;
end.

